//
// RegexReplace
//
// $Id: RegexReplace.cc,v 1.0 2000/08/16 13:01:00 aarmstrong Exp $
//
//

#include "RegexReplace.h"
#include <stream.h>

// Construct a RegexReplace that will translate strings that match
// the pattern in |from| with the replacement pattern in |to|. If pError
// is supplied it will contain any error condition that occurs while
// compiling the RE
RegexReplace::RegexReplace(const String &from, const String &to, String &error)
{
	memset(&re, 0, sizeof(re));
	memset(&regs, 0, sizeof(regs));
	repBuf		= 0;
	segSize		=
	segUsed		= 0;
	segMark		= 0;
	repLen		= 0;

	//toString = "[replace \"" + from + "\" with \"" + to + "\"]";

	re_set_syntax(RE_SYNTAX_AWK);

	const char *err = re_compile_pattern(from.get(), from.length(), &re);
	if (err != 0)
	{
		// Even though assignment to a String doesn't trash the source char *
		// there's no const char * variant of it. Hmm.
		error = (char *) err;
		return;
	}

	ParseReplaceSpec(to);			// can't generate an error
}

RegexReplace::~RegexReplace()
{
	Empty();
}

String RegexReplace::Replace(const String &source)
{
	int len = source.length();
	const char *src = source.get();
	//cout << "Replace(\"" << source.get() << "\")" << endl;
	if (re_match(&re, src, len, 0, &regs) == len)			// whole string matches?
	{
		// Firstly work out how long the result string will be. We think this will be more effecient
		// than letting the buffer grow in stages as we build the result, but who knows?
		//cout << "!!! Match !!!" << endl;
		size_t resLen = repLen;
		int i, reg, repPos;
		for (i = 1; i < segUsed; i += 2)
		{
			reg = segMark[i];
			if (reg < (int) regs.num_regs && regs.start[reg] != -1)
				resLen += regs.end[reg] - regs.start[reg];
		}
		//cout << "result will be " << resLen << " chars long" << endl;
		String result(resLen);	// Make the result string preallocating the buffer size
		for (i = 0, repPos = 0;; )
		{
			//cout << "appending segment " << i << endl;
			result.append(repBuf + repPos, segMark[i] - repPos);		// part of the replace string
			repPos = segMark[i];		// move forward
			if (++i == segUsed) break;	// was that the last segment?
			reg = segMark[i++];			// get the register number
			if (reg < (int) regs.num_regs && regs.start[reg] != -1)
				result.append((char *) src + regs.start[reg], regs.end[reg] - regs.start[reg]);
		}
		//cout << "return " << result.get() << endl;
		return result;
	}
	return source;
}

// Private helpers
void RegexReplace::Empty()
{
	// Hmm. What do we need to do about disposing of the workspace
	// that gets claimed for the regex?

    delete repBuf; repBuf = 0;
    segSize = segUsed = 0;
    delete segMark; segMark = 0;
    repLen = 0;
}

// Private: place a mark in the mark buffer growing it if necessary.
void RegexReplace::PutMark(int n)
{
	// assert(segUsed <= segSize);
	if (segUsed == segSize)
	{
		size_t newSize = segSize * 2 + 5;		// grow in chunks
		int *newMark = new int[newSize];		// do we assume that new can't fail?
		memcpy(newMark, segMark, segSize * sizeof(int));
		delete segMark;
		segMark = newMark;
		segSize = newSize;
	}
	segMark[segUsed++] = n;
}

void RegexReplace::ParseReplaceSpec(const String &to)
{
	const char *top = to.get();
	const char *end = top + to.length();

	repBuf = new char[to.length()];		// replace buffer can never contain more text than to string
	int bufPos = 0;			// our position within the output buffer

	while (top < end)
	{
		if (*top == '\\')
		{
			if (++top >= end) break;
			if (*top >= '0' && *top <= '9')
			{
				PutMark(bufPos);
				PutMark(*top - '0');
			}
			else
			{
				// We could handle some C style escapes here, but instead we just pass the character
				// after the backslash through. This means that \\, \" and \' will do the right thing.
				// It's unlikely that anyone will need any C style escapes in ht://Dig anyway.
				repBuf[bufPos++] = *top;
			}
			top++;
		}
		else
		{
			repBuf[bufPos++] = *top++;
		}
	}
	PutMark(bufPos);
	repLen = (size_t) bufPos;
}

// End of RegexReplace.cc
